#!/usr/bin/python3

from colored import fg, bg, attr
import os
import platform
import re
from time import time, sleep
from subprocess import Popen, PIPE
import evdev
from evdev import ecodes, InputDevice
import serial

class qwertyGroup:
    """ Categorizes certain keys into groups """

    esc = [ecodes.KEY_ESC]

    left = [
        ecodes.KEY_TAB,
        ecodes.KEY_CAPSLOCK,
        ecodes.KEY_LEFTSHIFT,
        ecodes.KEY_LEFTCTRL,
        ecodes.KEY_LEFTMETA,
        ecodes.KEY_LEFTALT]

    space = [ecodes.KEY_SPACE]

    right = [
        ecodes.KEY_COMPOSE,
        ecodes.KEY_RIGHTMETA,
        ecodes.KEY_RIGHTALT,
        ecodes.KEY_RIGHTCTRL,
        ecodes.KEY_RIGHTSHIFT,
        ecodes.KEY_ENTER,
        ecodes.KEY_BACKSLASH]

    arrow = [
        ecodes.KEY_UP,
        ecodes.KEY_DOWN,
        ecodes.KEY_LEFT,
        ecodes.KEY_RIGHT]

    numpad = [
        ecodes.KEY_KP0,
        ecodes.KEY_KPDOT,
        ecodes.KEY_KPENTER,
        ecodes.KEY_KP1,
        ecodes.KEY_KP2,
        ecodes.KEY_KP3,
        ecodes.KEY_KP4,
        ecodes.KEY_KP5,
        ecodes.KEY_KP6,
        ecodes.KEY_KPPLUS,
        ecodes.KEY_KP7,
        ecodes.KEY_KP8,
        ecodes.KEY_KP9,]
        #ecodes.KEY_KPSLASH,        Sticky keys on one keyboard
        #ecodes.KEY_KPASTERISK,
        #ecodes.KEY_KPMINUS,



###
def find_qwerty(name=''):
    print(name)
    p = re.compile('(?i)keyboard')
    for dev in [evdev.InputDevice(path) for path in evdev.list_devices()]:
        if p.search(dev.name):
            if name:
                if re.compile('(?i){}'.format(name)).search(dev.name):
                    return dev
            else:
                return dev

def find_mouse(name=''):
    # not both mice say "mouse"
    p = re.compile('(?i)m')
    for dev in [evdev.InputDevice(path) for path in evdev.list_devices()]:
        if p.search(dev.name):
            if name:
                if re.compile('(?i){}'.format(name)).search(dev.name):
                    return dev
            else:
                return dev
###

class qwertyStompbox:
    """ qwerty -> Stompbox
            For unrefined keypresses """
    def __init__(self, dev_name='', path=''):
        if path:
            self.dev = evdev.InputDevice(path)
        else:
            self.dev = self.find_qwerty(dev_name)
        self.set_led(0, 0)
        self.set_led(1, 0)
        self.set_led(2, 0)

        self.animate_leds()

    def find_qwerty(self, name=''):
        return find_qwerty(name)

    def set_led(self, index, state):
        self.dev.set_led(index, state)

    def leds(self):
        return self.dev.leds()


    def animate_leds(self):
        dt = 0.07
        frames = [
            [1,0,0],
            [1,1,0],
            [1,1,1],
            [0,1,1],
            [0,0,1],
            [0,0,0],
            [0,0,1],
            [0,1,1],
            [1,1,1],
            [1,1,0],
            [1,0,0],
            [0,0,0]]

        for frame in frames:
            ln = 0
            for i in range(3):
                self.set_led(i, frame[i])
            # sleep here is an obvious problem.
            # This will one day be event driven.
            sleep(dt)

    def listen(self):
        self.dev.grab()
        downs = []
        going = True

        # Blocking loop
        # self.dev is currently defined at top
        for event in self.dev.read_loop():
            if event.type == evdev.ecodes.EV_KEY:
                kcode = event.code
                estr = str(evdev.categorize(event))
                kconst = estr[estr.find('(')+1:estr.find(')')]
                letter = kconst[4:]
                if event.value == 1:
                    if kcode not in downs: # For some reason, I got a duplicate
                        downs.append(kcode)
                    # To stop, press Ctrl+c+Enter
                    # if kcode == ecodes.KEY_ENTER and ecodes.KEY_C in downs and (ecodes.KEY_LEFTCTRL in downs or ecodes.KEY_RIGHTCTRL in downs):
                    # Or, press Esc+Num_Pad_Enter
                    if kcode == ecodes.KEY_KPENTER and ecodes.KEY_ESC in downs:
                        self.stop()

                    for metabutton in self.metabuttons:
                        if kcode in metabutton.keys:
                            if not metabutton.on: # Turn on
                                metabutton.latch()
                                metabutton.press()
                                metabutton.on = True
                            metabutton.keydowns.append(kcode)

                if event.value == 0:
                    if kcode in downs: # In case modifiers report keydown again
                        try:
                            downs.remove(kcode)
                        except ValueError:
                            print(ValueError)
                    else:
                        print(kcode, ' not in downs')

                    # Recheck all keys that are down
                    for metabutton in self.metabuttons:
                        if len(metabutton.keydowns) > 0:
                            if kcode in metabutton.keys:
                                try:
                                    metabutton.keydowns.remove(kcode)
                                    metabutton.keydowns.remove(kcode)
                                except ValueError:
                                    pass
                            if len(metabutton.keydowns) == 0:
                                metabutton.on = False
                                metabutton.release()
                if event.value == 2:
                    for metabutton in self.metabuttons:
                        if metabutton.label == 'right':
                            if metabutton.on:
                                pass
                                #hog2volres()

                if event.value != 2:
                    #color_key_print()
                    #print(leds)
                    pass

    def stop(self):
        self.dev.ungrab()
        #animate_leds(self)
        self.animate_leds()
        exit()

class MetaButton:
    def __init__(self, label, keys, qwerty):
        self.label = label
        self.keys = keys
        self.qwerty = qwerty
        self.keydowns = []
        self.on = False
        self.switch = False
        
        self.active = True
        # This is for a normally on switch to turn it off.
        # This state is currently unknown by this class.
        # Instead, the escape class will set the led and send the midi

    def latch(self):
        self.switch = not self.switch

    def press(self):
        self.on = True

    def release(self):
        self.on = False

    def activate(self):
        self.active = True

    def deactivate(self):
        self.active = False


def send_midi(cmd, note, vel):
    if SERIAL:
        print(bytes([cmd, note, vel]))
        #ser.write(bytes([cmd, note, vel]))
        return

    elif SCREEN:
        """ Sends midi signal to arduino through serial. Baud rate is 31250.

            Set screen session with (gnu/linux)
                screen -S arduino /dev/ttyACM0 31250
            (osx)
                screen -S arduino /dev/cu.usbmodem411 31250

            Support is mostly good! Some characters are probably not properly 
            escaped! Pros are that you can monitor serial input through screen. 
            Also, other programs, such as the shell, can send midi signals as well.
            
            Examples:

                Send b0 61 7f (176 97 127) with
                    screen -S arduino -p 0 -X stuff $'\xb0\x61\x7f'
                Send b0 61 00 (176 97 00) with
                    screen -S arduino -p 0 -X stuff $'\xb0\x61^@'
        """
        midiarr = [cmd,note,vel]
        # This is dumb, bad, and wrong. 
        # You should escapbe by writing bash bytes with byte string
        # \\ x 6 1
        # like 
        # [ord(c) for c in '\\x' + hex(cmd)[2:4]]
        # which is 
        # bytes([92,120,54,49])
        # and if any are 0x00, ^@

        # Let's try this
        p = Popen('/bin/bash',stdout=PIPE,stderr=PIPE,stdin=PIPE)
        if vel == 0x00:
            out = p.communicate(b"screen -S arduino -p 0 -X stuff $'" + bytes([cmd, note]) + b"^@'")
        else:
            out = p.communicate(b"screen -S arduino -p 0 -X stuff $'" + bytes(midiarr) + b"'")

        ######
        return
        ######

        if cmd == 0xb0 and vel == 0x00:
            #print("Yep, that problem of properly escaping! Total hack: $'\\xb0\\x61^@")
            print("Yep, that problem of properly escaping! Total hack: $'" + 
                str(midiarr[:-1])[2:].replace('\\','\\\\') + "^@")
            #os.system(b"screen -S arduino -p 0 -X stuff $'\\xb0\\x61^@'")
            p = Popen('/bin/bash',stdout=PIPE,stderr=PIPE,stdin=PIPE)
            #out = p.communicate(b"screen -S arduino -p 0 -X stuff $'\\xb0\\x61^@'")
            out = p.communicate(b"screen -S arduino -p 0 -X stuff $'" + bytes([cmd, note]) + b"^@'")

            #os.system(b"screen -S arduino -p 0 -X stuff $'\\xb0\\x61^@'")
            return
        midiarr = [cmd,note,vel]
        i = 0
        while i < len(midiarr):
            if midiarr[i] in [ord(c) for c in '`\'"$ \\#[]!;{}~']:
                midiarr.insert(i, ord('\\'))
                i+=1
            i+=1
        if len(midiarr) > 3:
            print("Escaped midi string: {}".format(midiarr))
        else:
            os.system(b"screen -S arduino -p 0 -X stuff \"" + bytes(midiarr) + b'"')
    #if vel == 0x00:
    #    print("Yep, that problem of properly escaping! Total hack: $'" + 
    #        str(midiarr[:-1])[2:].replace('\\','\\\\') + "^@")
    #    p = Popen('/bin/bash',stdout=PIPE,stderr=PIPE,stdin=PIPE)
    #    out = p.communicate(b"screen -S arduino -p 0 -X stuff $'" + bytes([midiarr][:-1]) + b"^@'")
    #else:
    #    os.system(b"screen -S arduino -p 0 -X stuff \"" + bytes(midiarr) + b'"')

def hog2_spectral_gate(on):
    send_midi(0xb1,88,126*on+1) # 1, 127

ccs = range(1,80)
def hog2volres():
    global x, y, z, lastcci, adbpipe
    """ This method reads phone values through adb getevent """
    """ Because you don't want to send fake events to evdev to check for new key event,\
        this on is not a latch swith :.( """
    start = time()
    adbpipe.stdout.flush()
    while time() - start < 0.03:
        try:
            line = adbpipe.stdout.readline()
            l = str(line).split(' ')
            if len(l) > 3 and l[0].find('/dev/input/event4:') != -1:
            #if len(l) > 3 and l[0] == '/dev/input/event3:':
                zero, axis, mag = l[1:4]
                if zero == '0002':
                #if zero == '0003':
                    mag = int(mag[:mag.find('\\')], 16)
                    if mag > int('0xffff0000', 16):
                        mag = mag - 16**8
                    if axis == '0000':
                        x = mag
                        # this note will be from 20 - 80 (60 + 20)
                        # from -5000 to 5000, 0 = 30 + 20
                        # 5000 is 60 + 20, -5000 is 0 + 20

                        # Now lets instead choose 21 notes from a pentatonic scale
                        # -7000 to 7000, -7000 is 0, 7000 is 20
                        thiscci = int((mag + 8000) * (len(ccs) - 1) / 16000.0)
                        #thisnote = int((mag + 700) * (len(notes) - 1) / 1400.0)
                        if thiscci > len(ccs) - 1:
                            thiscci = len(ccs) - 1
                        elif thiscci < 0:
                            thiscci = 0
                        #thisnote = int(mag * 30.0 / 5000) + 70
                        
                        #print(x, '\t')
                        if lastcci != thiscci:
                        #    scm.send_midi((0x90, lastcci, 0x00))
                        #    scm.send_midi((0x90, thiscci, 0x50))
                            send_midi(0xb0, 0x73, ccs[thiscci])
                            #scm.send_midi((0xe0, 0x00, ccs[thiscci]))
                            lastcci = thiscci

                            #print(thiscci,lastcci)
                    if axis == '0001':
                        y = mag
                    if axis == '0002':
                        z = mag
                #print(zero, axis, mag, flush=True)
                #else:
                    #pass
                    #print(x,y,z,),
        except Exception as exc:
            print(exc)

def color_key_print():
    # A nice to have someday
    return

SERIAL = 1
if __name__ == '__main__':
    global kbdleds, x, y, z, lastcci, adbpipe
    kbdleds = 0b000
    x = y = z = lastcci = 0
    SERIAL = 1
    SCREEN = 0
    if SERIAL:
        arduino_path = '/dev/ttyUSB0'
        #ser = serial.Serial(arduino_path, 57600)
    #adbpipe = Popen(['adb','shell','getevent'], stdin=PIPE, stdout=PIPE)
    #go()

