import sys, mido
import inspect
import logging
from xenedoji import Pitch
pitch = Pitch
from miface import MIdevice

#logging.basicConfig(level=logging.INFO)
logging.basicConfig(filename='log/prelude.log', filemode='w', level=logging.INFO)

              # Change to MDevice
class Keyboard(MIdevice):
    # This class will only accept on/off messages
    # Touch may be added later here or another class like TouchKeyboard
    def __init__(self, bottom_key, top_key, name, midi_in=True, unused='', polyphonic=True):
        # changed bottom_key to a pitch
        if isinstance(bottom_key, str):
            self.bottom_key = Pitch(bottom_key)
        elif isinstance(bottom_key, int):
            self.bottom_key = Pitch(bottom_key)
        elif isinstance(bottom_key, Pitch):
            self.bottom_key = bottom_key
        if isinstance(top_key, str):
            self.top_key = Pitch(top_key)
        elif isinstance(top_key, int):
            self.top_key = Pitch(top_key)
        elif isinstance(top_key, Pitch):
            self.top_key = top_key
        self.name = name
        self.midi_in = midi_in # Only True for now
        self.polyphonic = polyphonic
        if self.midi_in:
            self.iport = mido.open_input(self.name, virtual=True)
            self.tport = mido.open_output('{} Thru'.format(self.name), virtual=True)
        self.connect_to_monitor()
        #self.iport = mido.open_input('Pedalboard')
        #self.iport.callback = self.interpret

        # self.range?
        # self.len or self.key_len
        # maybe change to self.keys with an s
        self.key = []
        for i in range(self.bottom_key.midi(), self.top_key.midi() + 1):
        #for i in range(self.bottom_key.midi() - self.top_key.midi() + 1):
            #self.key.append(MKey(i, Pitch(i) + (self.bottom_key - Pitch(0)), self))
            self.key.append(MKey(i, Pitch(i), self))
        self.keys_on = []
        self.callback = None
        # These may be later settable by a funciton
        self.note_on_callback = None
        self.note_off_callback = None
        self.sustain_on = False


    def send(self, msg):
        self.interpret(msg)
    # Should be called set_key_callback or set_key_onoff_callback
    def setcallback(self, callback):
        # Notes that are currently on will get old note off if found
        for key in self.keys_on:
            # What is this mess? You have to clean this up
            if not (key.note_off_queue() or key.note_off_callback or key.callback):
                if self.note_off_callback:
                    key.note_off_queue(self.note_off_callback)
                else:
                    key.note_off_queue(self.callback)
        # Should be called _key_callback or _key_onoff_callback
        self.callback = callback

    def interpret(self, msg):
        if msg.type == 'note_on' or msg.type == 'note_off':
            if not self.polyphonic and self.is_on():
                if len(self.keys_on) > 1:
                    logging.error('More than one key down for monophonic keyboard')
                if len(self.keys_on) == 1 and msg.type == 'note_off' and True: # if not note_off, keyboard_off
                    self.from_midi(msg.note).interpret(msg)
                elif len(self.keys_on) == 1 and msg.type == 'note_on':
                    for key in self.keys_on:
                        if key.midi != msg.note:
                            key.off()
                #for key in self.keys_on: # Assuming more than one would be on 
                #    if key.midi != msg.note: # for some reason
                #        key.off()
            self.from_midi(msg.note).interpret(msg)
        elif msg.is_cc(0x40):
            if msg.value >= 0x40:
                self.on_sustain_on()
            else:
                self.on_sustain_off()

    def on_keyboard_on(self, msg):
        pass
    def on_keyboard_off(self, msg):
        pass
    def is_on(self):
        return bool(self.keys_on)

    def on_sustain_on(self):
        self.sustain_on = True
        if self.sustain_on_callback:
            self.sustain_on_callback()
    def on_sustain_off(self):
        self.sustain_on = False
        if self.sustain_off_callback:
            self.sustain_off_callback()

    def from_midi(self, key):
        return self.key[key - self.bottom_key.midi()]

    def all_notes_off(self):
        for note in self.keys_on:
            note.off()

    #def set_callback(self, callback)
    #   This can be used later to queue callbacks until

    def __len__(self):
        return len(self.key)

    def __getitem__(self, index):
        if isinstance(index, str):
            index = Pitch(index)
        if isinstance(index, pitch):
            #pitch.midi
            return self.from_midi(index.midi())
        elif isinstance(index, int):
            return self.key[index]

class MKey:
    def __init__(self, index, pitch, keyboard):
        self.index = index
        self.pitch = pitch
        self.midi = pitch.midi()
        self.keyboard = keyboard
        # ignore secondary note ons before note off
        self.polyphonic = True
        # This is to track extra key on/off messages
        self.on_position = 0
        self.off_position = 1
        self.is_on = False
        self._note_on_queue = []
        self._note_off_queue = []
        self.note_on_callback = None
        self.note_off_callback = None
        self.callback = None

    def on(self, vel=64):
        # This index thing has to be fixed
        #self.on_note_on(mido.Message('note_on', note=self.index+self.keyboard.bottom_key.midi(), velocity=vel))
        self.on_note_on(mido.Message('note_on', note=self.index, velocity=vel))

    def off(self, msg=None):
        if msg:
            self.on_note_off(msg)
        else:
            # This index thing has to be fixed
            #self.on_note_off(mido.Message('note_off', note=self.index+self.keyboard.bottom_key.midi()))
            self.on_note_off(mido.Message('note_off', note=self.index))
        # bottom_key could be a pitch instead? Yes

    def setcallback(self, function):
        if self.is_on:
            if self._note_off_queue:
                # Should this be here?
                # If a note off function is already queued, don't worry about it
                pass
            elif self.callback:
                self.note_off_queue(self.callback)
            # Should be called keyboard._key_callback or keyboard._key_onoff_callback
            elif self.keyboard.callback:
                self.note_off_queue(self.keyboard.callback)
        self.callback = function

    def queue(self, function):
        self._note_on_queue.append(function)
        self._note_off_queue.append(function)

    def note_on_queue(self, function=None):
        if function:
            self._note_on_queue.append(function)
        return self._note_on_queue

    def note_off_queue(self, function=None):
        # This function exists because I don't want to use append
        # note_off_queue.append is too many words
        if function:
            self._note_off_queue.append(function)
        return self._note_off_queue

    def on_note_on(self, msg):
        #print('{}\t{}'.format(str(self.pitch),str(msg)))
        self.is_on = True
        self.on_position += 1
        self.off_position -= 1
        if self.on_position > 1 and not self.polyphonic: # Polyphonic here
                # only applys to this note, not others (keyboard.polyphonic).
            #print('Error, multiple note on\'s sent to monophonic key')
            logging.warning('Multiple note on\'s sent to monophonic key {}'.format(self.pitch.spn()))
        else:
            note_on_callback = None
            if self.keyboard.is_on():
                self.keyboard.on_keyboard_on(msg) # This name is bad
            self.keyboard.keys_on.append(self)
            if self._note_on_queue:
                note_on_callback = self._note_on_queue.pop(0)
                logstr = '{} note on. Using _note_on_queue: '.format(self.pitch)
                logstr += inspect.getsource(note_on_callback)
                logging.info(logstr)
            elif self.note_on_callback:
                note_on_callback = self.note_on_callback
                logstr = '{} note on. Using _note_on_callback: '.format(self.pitch)
                logstr += inspect.getsource(note_on_callback)
                logging.info(logstr)
            elif self.callback:
                note_on_callback = self.callback
                logstr = '{} note on. Using self.callback: '.format(self.pitch)
                logstr += inspect.getsource(note_on_callback)
                logging.info(logstr)
            elif self.keyboard.callback:
                note_on_callback = self.keyboard.callback
                logstr = '{} note on. Using keyboard.callback: '.format(self.pitch)
                logstr += inspect.getsource(note_on_callback)
                logging.info(logstr)

            if note_on_callback:
                #logging.info(inspect.get_source(note_on_callback))
                note_on_callback(msg)

    def on_note_off(self, msg):
        #print('{}\t{}'.format(str(self.pitch),str(msg)))
        self.is_on = False
        self.on_position -= 1
        self.off_position += 1
        if self.on_position > 1 and not self.polyphonic:
            logging.warning('Multiple note off\'s sent to monophonic key {}'.format(self.pitch.spn()))
        else:
            try:
                self.keyboard.keys_on.remove(self)
            except ValueError:
                print("{} not in keyboard.keys_on".format(self.pitch))
            if self.keyboard.is_on():
                self.keyboard.on_keyboard_off(msg) # This name is bad
        note_off_callback = None
        if self._note_off_queue:
            note_off_callback = self._note_off_queue.pop(0)
            logstr = '{} note off. Using _note_off_queue: '.format(self.pitch)
            logstr += inspect.getsource(note_off_callback)
            logging.info(logstr)
        elif self.note_off_callback:
            note_off_callback = self.note_off_callback
            logstr = '{} note off. Using _note_off_callback: '.format(self.pitch)
            logstr += inspect.getsource(note_off_callback)
            logging.info(logstr)
        elif self.callback:
            note_off_callback = self.callback
            logstr = '{} note off. Using self.callback: '.format(self.pitch)
            logstr += inspect.getsource(note_off_callback)
            logging.info(logstr)
        elif self.keyboard.callback:
            note_off_callback = self.keyboard.callback
            logstr = '{} note off. Using keyboard.callback: '.format(self.pitch)
            logstr += inspect.getsource(note_off_callback)
            logging.info(logstr)

        if note_off_callback:
            note_off_callback(msg)

    def interpret(self, msg):
        if msg.type == 'note_on':
            self.on_note_on(msg)
        elif msg.type == 'note_off':
            self.on_note_off(msg)

#k = Keyboard('C1', 'C3')

#MMsg() | MStatus(9) | MChannel(1) | Pitch('A2') | MVel(50)
#MMsg() | MStatus(9) | MChannel(1) | MKey('A2') | Dynamic('mf')

#print(inspect.getsource(keyboard.callback))

def display_callbacks(keyboard):
    tree = ''
    if keyboard.callback:
        tree += inspect.getsource(keyboard.callback)
    if keyboard.note_on_callback:
        tree += ' '*2 + inspect.getsource(keyboard.note_on_callback)
    if keyboard.note_off_callback:
        tree += ' '*2 + inspect.getsource(keyboard.note_off_callback)
    for key in keyboard.key:
        branch = ''
        if key.callback:
            branch += ' '*4 + inspect.getsource(key.callback)
        if key.note_on_callback:
            branch += ' '*6 + inspect.getsource(key.note_on_callback)
        if key.note_off_callback:
            branch += ' '*6 + inspect.getsource(key.note_off_callback)
        on_queue = key.note_on_queue()
        try:
            for i in range(len(on_queue)):
                branch += ' '*8 + inspect.getsource(on_queue[i])
        except IndexError:
            # Multiple times program has crashed because callback has been
                # removed between checking length and indexing last
            pass
        off_queue = key.note_off_queue()
        try:
            for i in range(len(off_queue)):
                branch += ' '*8 + inspect.getsource(off_queue[i])
        except IndexError:
            # Multiple tives program has crashed because callback has been
                # removed between checking length and indexing last
            pass
        if branch:
            tree += '  {}:\n{}'.format(key.pitch, branch)
    return tree

#keyboard[3].callback = lambda msg: print('keyboard[3].callback')
#keyboard[4].setcallback(lambda msg: print('keyboard[4].callback'))
#keyboard[5].note_off_queue(lambda msg: print('keyboard[5].note_off_queue'))
#print(display_callbacks(keyboard))
