/*
 * Button press sends a sustain message to connected device and
 * computer. It also ligths the led blue and green.
 * 
 * Green flash for midi in events.
 * Blue flash for midi RX events.
 * 
 * Midi mod wheel sets the red led.
 */

#include <MIDIVIL.h>

bool query_midpoint = false;
int midpoint = 62;
bool negate = false;

void set_negate(bool negate_status);
void allNotesOff();
bool is_negate();

void setup() {
  Serial.begin(57600);
  pinMode(Rpin, OUTPUT); 
  pinMode(Gpin, OUTPUT);
  pinMode(Bpin, OUTPUT);
  pinMode(Btnpin, INPUT);
  midiSerial.begin(31250);
  midiSerial1.begin(31250);

  setLed(0,0,0);
  //flash(511,511,511);
}

void loop() {
  buttonEventController();
  midiEventController();
  updateLed();
}

void updateLed() {
  if (flashing && millis() > flash_start + flash_interval) {
    flashing = false;
    //setLed(1023-rgb[0],1023-rgb[1],1023-rgb[2]);
    ///*
    if (isPressed()) {
      setLed(rgb[0],1023,1023);
    } else {
      setLed(rgb[0],0,rgb[2]);
    }//*/
  }
}

void setLed(int R, int G, int B) {
  analogWrite(Rpin, 1023-R);
  analogWrite(Gpin, 1023-G);
  analogWrite(Bpin, 1023-B);
  rgb[0] = R;
  rgb[1] = G;
  rgb[2] = B;
}

void onPress() {
  //Serial.println("onPress");
  //allNotesOff();
}

void onMidiIn(int b0, int b1, int b2) {
  //printMidi(b0, b1, b2);
  //printMidi(b0, midpoint +60 - b1, b2);
  //return;
  // this code allows an extra keyup to pass after query_midpoint
  if (query_midpoint and b0 == 128 /* keydown */ ) {
    midpoint = b1;
    query_midpoint = false;
    //flash(0,511,0);
    // Turn on negative piano after key is detected
    set_negate(true);
  } else if (query_midpoint == false) {
    if (is_negate()) {
      //sendMidi(b0, midpoint - b1, b2);
      printMidi(b0, midpoint - (b1-midpoint), b2);
    } else {
      //sendMidi(b0, b1, b2);
      printMidi(b0, b1, b2);
    }
    if (b0 == 144) {
      //allNotesOff();
    }
    flash(rgb[0],511,rgb[2]);
  }
}

void allNotesOff() {
  // This a bad way to get around a stuck key problem I am having
  // only occasionally
  for (int i = 20; i < 100; i++) {
    sendMidi(0x90, i, 0);
    printMidi(0x90, i, 0);
  }
}

void onRelease() {
  //Serial.println("onRelease");
  // Do nothing, instead wait for keypress
}

void onLatchOn() {
  //Serial.println("onLatchOn");
  set_negate(true);
  //setLed(511,0,511); // Purple
}
void onLatchOff() {
  //Serial.println("onLatchOff");
  set_negate(false);
  //setLed(0,0,0);
}
void onHoldDown() {
  //Serial.println("onHoldDown");
  setLed(511,511,0); // Yellow
  query_midpoint = true;
}
void onHoldRelease() {
  //Serial.println("onHoldRelease");
}

// Called onPress of second tap
void onDTapPress() {
  //Serial.println("onDTapPress");
  //setLed(511,0,511); // Purple
}
// Called onRelease of second tap
void onDTapRelease() {
  //Serial.println("onDTapRelease");
}

// Latching Switches
void onDTapLatchOn() {
  //Serial.println("onDTapLatchOn");
}
void onDTapLatchOff() {
  //Serial.println("onDTapLatchOff");
}

void set_negate(bool negate_status) {
  negate = negate_status;
  if (negate) {
    setLed(511,0,511);
  } else {
    setLed(0,0,0);
  }
  //Serial.println(negate);
}

bool is_negate() {
  return negate;
}
