/*
 * Button press sends a sustain message to connected device and
 * computer. It also ligths the led blue and green.
 * 
 * Green flash for midi in events.
 * Blue flash for midi RX events.
 * 
 * Midi mod wheel sets the red led.
 */

#include "MIDIVIL.h"
#include <Arduino.h>

#include <SoftwareSerial.h>

SoftwareSerial midiSerial(13, 14);
SoftwareSerial midiSerial1(16, 12);

int Rpin = 2;
int Gpin = 0;
int Bpin = 4;
int RGBpin[3] = {Rpin, Gpin, Bpin};
int Btnpin = 5;
int MIpin = 14;
int MOpin = 12;

int state = 0;
int pressed = 0;
int press_start = 0;
int latch = 0; // bool

int dtap = 0; // bool
int tap = 0;
int tap_start = 0;

int dtap_interval = 400;

int hold_start = 0;
int hold_time = 0;
int held = 0; // bool
int holdlatch = 0; // bool
int hold_interval = 1000;

// This is for buttons that don't release cleanly
// I will not get such a switch for the thing
int jitter_interval = 20;
int jitter_start = 0;
// This would actually be better if it checked for three events
// in the window.
int jitter_check = 0; // Currently unused
////

void setLed(int R, int G, int B);
void buttonEventController();
void onMidiIn(int b0, int b1, int b2); /* Needed? */
void onPress();
void onRelease();
void onHoldDown();
void onHoldRelease();
void onLatchOn();
void onLatchOff();
// Called onPress of second tap
void onDTapPress();
// Called onRelease of second tap
void onDTapRelease();
// Latching Switches
void onDTapLatchOn();
void onDTapLatchOff();
bool isPressed();
bool isLatchOn();
bool isDTapLatchOn();

unsigned int rgbColor[3];
void updateLed();

int decColor = 0;
int incColor;
int color_step = 0;

bool flashing = false;
int flash_start = 0;
int flashColor[3] = {0,0,0};
int flash_interval = 200;
void flash(int r, int g, int b);
unsigned int rgb[3] = {0,0,0};

int timeout = 3; // sometimes get 0xffffffff
//void setLed(int R, int G, int B);
void sendMidi(int cmd, int key, int vel);
void printMidi(int cmd, int note, int vel);


void flash(int r, int g, int b) {
  flashColor[0] = r;
  flashColor[1] = g;
  flashColor[2] = b;
  setLed(flashColor[0],flashColor[1],flashColor[2]);
  flashing = true;
  flash_start = millis();
}

// Why did I comment this out?
///*
void setLed(int R, int G, int B) {
  analogWrite(Rpin, 1023-R);
  analogWrite(Gpin, 1023-G);
  analogWrite(Bpin, 1023-B);
  rgb[0] = R;
  rgb[1] = G;
  rgb[2] = B;
}
//*/

bool isPressed() {
  return pressed;
}
bool isLatchOn() {
  return latch;
}
bool isHeld() {
  return held;
}
bool isHoldLatchOn() {
  return holdlatch;
}
bool isDTapLatchOn() {
  return dtap;
}

void sendMidi(int cmd, int key, int vel) {
  if (cmd % 16 == 1) {
    midiSerial1.write(cmd);
    midiSerial1.write(key);
    midiSerial1.write(vel);
  } else {
    midiSerial.write(cmd);
    midiSerial.write(key);
    midiSerial.write(vel);
  }
}

void printMidi(int cmd, int note, int vel) {
  if (cmd < 16 && cmd != -1) { Serial.print("0"); }
  Serial.print(cmd, HEX);
  Serial.print(" ");
  if (note < 16 && note != -1) { Serial.print("0"); }
  Serial.print(note, HEX);
  Serial.print(" ");
  if (vel < 16 && vel != -1) { Serial.print("0"); }
  Serial.println(vel, HEX);
}

void buttonEventController() {
  int pressed_now = !digitalRead(Btnpin);

  if (pressed != pressed_now) {
    pressed = pressed_now;
    // This is where you chould use the variable "jitter_check"
    // To save the press or release.
    // If no second press or release is detected during the window
    // Let it through.
    // Is this too much?
    if (millis() > jitter_start + jitter_interval) {
      if (pressed) {
        press_start = millis();
        latch = (latch + 1) % 2;
        if (tap == 0) {
          tap = 1;
          tap_start = millis();
        } else if (tap == 1 && millis() < tap_start + dtap_interval) {
          tap = 2;
          onDTapPress();
        }
  
        // This is out of order.
        // All Vars should be set before function calls.
        onPress();
        if (latch) {
          onLatchOn();
        } else {
          onLatchOff();
        }
        
      } else {
        // Release
        if (tap == 2 && millis() < tap_start + dtap_interval) {
          dtap = (dtap + 1) % 2;
          onDTapRelease();
          // Latching Double Tap Switch
          if (dtap) {
            onDTapLatchOn();
          } else {
            onDTapLatchOff();
          }
          tap = 0;
        }
        
        // Out of Order
        onRelease();
        if (held) {
          held = 0;
          hold_time = 0;
          onHoldRelease();
        }
      }
    } else {
      //Serial.println("Jitter Protection!");
    }
    jitter_start = millis();
  }
  if (tap > 0) {
    if (millis() > tap_start + dtap_interval) {
     tap = 0;
    }
  }
  if (pressed) {
    hold_time = millis() - press_start;
    if (millis() > press_start + hold_interval) {
      if (!held) {
        held = 1;
        onHoldDown();
        holdlatch = (holdlatch + 1) % 2;
      }
    }
  }
}

void midiEventController() {
  // This Function should be in one place, not every file!
  // And it should work better.

  // Midi In to TX
  int avail = midiSerial.available();
  if (avail > 0) {
    int cmd = midiSerial.read();
    if (cmd != 254) { //cmd == 128 || cmd == 144 || cmd == 176) {
      avail = midiSerial.available();
      int midi_start = millis();      
      while (avail < 1 && millis() < midi_start + timeout) {
        avail = midiSerial.available();
      }
      int note = midiSerial.read();
      avail = midiSerial.available();
      midi_start = millis();
      while (avail < 1 && millis() < midi_start + timeout) {
        avail = midiSerial.available();
      }
      int vel = midiSerial.read();
      /*
      if (cmd == 0xb0 && note == 0x01) {
        //printMidi(cmd,note,vel);
        // This should be in an onMidiIn(b0,b1,b2)
        onMidiIn(cmd,note,vel);

        setLed(vel*5,rgb[1],rgb[2]);
      } else {
        //printMidi(cmd,note,vel);
        onMidiIn(cmd,note,vel);
        // only applies to rx as of now
        flash(rgb[0],1023,0);
      }*/
      onMidiIn(cmd,note,vel);
    }
  }

  // RX to Midi Out
  avail = Serial.available();
  if (avail > 0) {
    int cmd = Serial.read();
    if (cmd != 254) { //cmd == 128 || cmd == 144 || cmd == 176) {
      avail = Serial.available();
      int midi_start = millis(); 
      //Serial.print("avail: "); Serial.println(avail);    
      while (avail < 1 && millis() < midi_start + timeout) {
        avail = Serial.available();
        //Serial.print("avail: "); Serial.println(avail); 
      }
      int note = Serial.read();
      // Sometimes get 0xffffffff

      // This Sucks
      while (note == -1) {
        //midi_start = millis(); 
        while (avail < 1 && millis() < midi_start + timeout) {
          avail = Serial.available();
          //Serial.print("avail: "); Serial.println(avail); 
        }
        note = Serial.read();
      }
      
      avail = Serial.available();
      midi_start = millis();      
      while (avail < 1 && millis() < midi_start + timeout) {
        avail = Serial.available();
      }
      int vel = Serial.read();

      // This Bad
      while (vel == -1) {
        //midi_start = millis();      
        while (avail < 1 && millis() < midi_start + timeout) {
          avail = Serial.available();
        }
        vel = Serial.read();
      }

      sendMidi(cmd,note,vel);
      //printMidi(cmd,note,vel);
      if (cmd != 0xb0 || note != 0x01) {
        flash(rgb[0],0,1023);
      }
    }
  }
}
