from math import log, sqrt

class Edostep:
    def __init__(self, step, edo):
        self.step = step
        self.edo = edo
        if isinstance(self.step, int) and divisible_by(step * 1200, edo):
            self.cents = Cents(step * 1200 // edo)
        else:
            self.cents = Cents(step * 1200 / edo)

    def reduce(self):
        if isinstance(self.step, int) and isinstance(self.edo, int):
            r = Ratio(self.step, self.edo)
            self.step = r.numerator
            self.edo = r.denominator
            return self
        else:
            print('step and edo are not both ints')

    def __repr__(self):
        return '{}\\{}'.format(self.step, self.edo)

def divisible_by(n, d):
    return Ratio(n, d).denominator == 1

USE_SUBSCRIPT = False
subscript = dict.fromkeys([str(i) for i in range(10)]+['.','-'])
for i in range(10):
    subscript[str(i)] = chr(0x2080 + i)
subscript['-'] = '\u208b'
subscript['.'] = '.'
number = {v: k for k, v in subscript.items()}

def number_str_to_subscript(n):
    return ''.join([subscript[c] for c in str(n)])

def subscript_to_number_str(substr):
    return ''.join([number[c] for c in substr])

USE_UNICODE_ACCIDENTALS = False
unicode_accidental = {
    '\u266d': '\u266d',
    'b': '\u266d',
    ' ': ' ',
    '#': '\u266f',
    '\u266f': '\u266f'
}

class Pitch:
    def __init__(self, midi_key, cents=0):
        if type(midi_key) == str:
            # I know this is bad!
            self.midi_key = Pitch.from_name(midi_key).midi_key
            self.cents = Cents(cents)
        elif type(midi_key) == float:
            self.midi_key = int(midi_key)
            self.cents = Cents((midi_key % 1) * 100) + cents
            #self.cents = Cents((midi_key % 1) * 100) + Cents(cents)
            # but not
            #self.cents = cents + Cents((midi_key % 1) * 100)
            self.is_12_edo = False
        elif type(midi_key) == int:
            self.midi_key = midi_key
            self.cents = Cents(cents)
        elif type(midi_key) == Hz:
            self.freq = midi_key
            midi_key = self.freq.to_midi_key()
            self.cents = Cents((midi_key % 1) * 100) + cents
            self.midi_key = int(midi_key)
        else:
            print("Exception")

        self.round_cents()
        self.set_name()
        self.set_freq() # This is unnecessary if self.freq is passed

    def midi(self):
        return self.midi_key

    def name(self):
        return self.spn()

    def spn(self):
        if USE_SUBSCRIPT:
            i = number_str_to_subscript(self.note_name_i)
        else:
            i = str(self.note_name_i)
        if USE_UNICODE_ACCIDENTALS:
            accidental = unicode_accidental[self.note_accidental].strip()
        else:
            accidental = self.note_accidental.strip()
        return self.note_name_base + accidental + i
        #return self.note_name

    def round_cents(self):
        # 100 c and -100c don't work

        # round_cents is not a good name
        # This method rounds the note and cents to the nearest note
        if int((self.cents//50).value) >= 0:
            self.midi_key = int((self.cents//50).value) + self.midi_key
            # += is done through iadd, this is why Cents.__add__ was not called
        else:
            self.midi_key = 1 + int((self.cents//50).value) + self.midi_key
        self.cents = Cents( (self.cents+50)%100 - 50 )

    def set_name(self):
        #if self.midi_key in range(100): # These are integers
        self.note_name_i = self.midi_key // 12 - 1
        #sharp_names = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
        sharp_accidentals_str = ' # #  # # # '
        sharp_names_str       = 'CCDDEFFGGAAB'
        i = int(self.midi_key % 12) # Why is this float? Remove int later
        self.note_name_base = sharp_names_str[i]
        self.note_accidental = sharp_accidentals_str[i]

    def from_name(name):
        """ Clas method as is used now? """
        """ How should this work? Where should this go? Should it return """
        """ a new object?"""
        #if len(name) < 2:
        #    """ Exception """
        #if len(name) > 2
        #    """ Exception """
        sharp_names_str       = 'CCDDEFFGGAAB'
        midi_key = 0
        if name[0].isalpha():
            base_name = name[0]
            if name[1] in ['b', '#', chr(9837), chr(9838), chr(9839)]:
                # And of course I don't handle anything besides # and b 
                #   for now
                # Double sharps and flats at 119082 and 119083 btw
                accidental = name[1]
                if name[2].isdigit():
                    if name[2].isdecimal():
                        octave = int(name[2])
                    elif name[2] in number.keys():
                        octave = int(subscript_to_number_str(name[2]))
                else:
                    """ Exception """
                    pass
            elif name[1].isdigit():
                """ Negative numbers not yet included """
                accidental = None
                if name[1].isdecimal():
                    octave = int(name[1])
                elif name[1] in number.keys():
                    octave = int(subscript_to_number_str(name[1]))

            midi_key = sharp_names_str.find(base_name)
            if accidental in ['b', '\u266d']:
                midi_key -= 1
            if accidental in [None, chr(9840)]:
                """ Well, I guess do nothing """
                pass
            if accidental in ['#', '\u266f']:
                midi_key += 1
            octave += midi_key//12
            # C0 is 12
            midi_key += octave * 12 + 12
            #print(midi_key)
            return Pitch(midi_key)
            
        else:
            print(""" Exception """)
            pass

    def accidental(self):
        return self.accidental

    def set_freq(self):
        #    self.freq = 440 * 2**((self.midi_key-49 + (self.cents/100.0).value) / 12)
        #TypeError: unsupported operand type(s) for +: 'int' and 'instance'
        self.freq = Hz(440 * 2**((self.midi_key-69 + self.cents.value/100.0) / 12))

    def get_freq(self):
        # freq(self) would be cool
        return self.freq

    def __eq__(self, other):
        return self.midi_key == other.midi_key

    def __gt__(self, other):
        return self.midi_key > other.midi_key

    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)

    def __lt__(self, other):
        return self.midi_key < other.midi_key

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __add__(self, other):
        if isinstance(other, Interval):
            return Pitch(self.midi_key + other.midi_interval * other.direction)
        elif isinstance(other, Cents):
            return Pitch(self.midi_key, self.cents + other.value)
        elif isinstance(other, Edostep):
            return Pitch(self.midi_key, self.cents + other.cents)
        #elif isinstance(other, Midi):
        #elif isinstance(other, Freq):

    def __sub__(self, other):
        if isinstance(other, Pitch):
            return Interval(self, other)
        elif isinstance(other, Interval):
            return self.__add__(-other)
        elif isinstance(other, Cents):
            return Pitch(self.midi_key, self.cents - other.value)
        elif isinstance(other, Edostep):
            return Pitch(self.midi_key, self.cents - other.cents)

    def __mul__(self, other):
        #print('%r * %r' % (self, other))
        if isinstance(other, float) or isinstance(other, int):
            # Calculate cents for now. Later will go by freq.
            p = 69 + 12*log2(self.get_freq() * other / 440.0)
            return Pitch(p)
        elif isinstance(other, Ratio):
            p = 69 + 12*log2(self.get_freq() * other.numerator / other.denominator / 440.0)
            return Pitch(p)
        "Calculate Freq, then get pitch"

    def __truediv__(self, other):
        return self.__div__(other)

    def __div__(self, other):
        #print('%r / %r' % (self, other))
        if isinstance(other, float) or isinstance(other, int):
            return self.__mul__(1 / other)
        elif isinstance(other, Ratio):
            return self.__mul__(other**-1)

    def __str__(self):
        if self.cents.value == 0:
            return self.spn()
        elif self.cents.value > 0:
            return self.spn() + ' +' + repr(self.cents)
        else:
            return self.spn() + ' ' + repr(self.cents)
        # This is where sign is needed

    def __repr__(self):
        return self.__str__()

    def get_cents(self):
        pass

"Pitch(Hz(440)"
"Pitch(A4)"
"Pitch(69)"
#class MidiNum
#class PitchBend
#class Freq
# or class Hz
# The plan is to get frequencies from 41edo pitch class
# Turn into 12edo +/- cents
# 

class Hz(float):
    def to_midi_key(self):
        return 12 * log2(self.real / 440) + 69

    def __add__(self, other):
        if not isinstance(other, Cents):
            return Hz(super().__add__(other))
        return Hz(self.real * 2**(other.value / 1200.0))

    def __sub__(self, other):
        if not isinstance(other, Cents):
            return Hz(super().__sub__(other))
        return Hz(self.real / 2**(other.value / 1200.0))

    def __neg__(self):
        return Hz(-self.real)

    def __repr__(self):
        if self.real.is_integer():
            return str(int(self.real)) + ' Hz'
        elif (self.real * 10).is_integer():
            #return ('%.1f¢') % self.real
            return str(self.real) + ' Hz'
            # I think these are the same. Check Cents too
        else:
            return ('%.2f Hz') % self.real
        #return '{:.2f} Hz'.format(self.real)

#class Hz: #(Freq):
#   def __init__(self, value):
#       self.value = value
#   def __repr__(self):
#       return '{} Hz'.format(self.value)
#   def __eq__(self, other):
#       return self.value == other.value

#   def __gt__(self, other):
#       return self.value > other.value

#   def __ge__(self, other):
#       return self.__gt__(other) or self.__eq__(other)

#   def __lt__(self, other):
#       return self.value < other.value

#   def __le__(self, other):
#       return self.__lt__(other) or self.__eq__(other)

#   def __add__(self, other):
#       return Hz(self.value + other.value)

#   def __sub__(self, other):
#       return Hz(self.value - other.value)

#   def __neg__(self):
#       return Hz(-self.value)

#   def __mul__(self, other):
#       if isinstance(other, int):
#           return Hz(self.value * other)

#   def __truediv__(self, other):
#       if isinstance(other, int):
#           return Hz(self.value / other)

class Cents: #(same superclass as Edostep uses?): or maybe (float):
    def __init__(self, value, number_of_decimal_places=2):
        if isinstance(value, self.__class__):
            self.value = value.value
            #self = value
            # Why can't I do that?
            self.number_of_decimal_places = value.number_of_decimal_places
        else:
            self.value = value
            self.number_of_decimal_places = number_of_decimal_places
            

    def to_cents(self, cents):
        if isinstance(cents, self.__class__):
            return cents
        if isinstance(cents, int) or isinstance(cents, float):
            return Cents(cents)

    def __add__(self, other):
        return Cents(self.value + self.to_cents(other).value)
    def __sub__(self, other):
        if isinstance(other, int):
            return Cents(self.value - self.to_cents(other).value)
        elif isinstance(other, Pitch):
            return Pitch.__add__(other, self)
    def __neg__(self):
        return Cents(-self.value)
    def __truediv__(self, other):
        return Cents(self.value / self.to_cents(other).value)
    def __mod__(self, other):
        return Cents(self.value % self.to_cents(other).value)
    def __floordiv__(self, other):
        return Cents(self.value // self.to_cents(other).value)
    
    def __repr__(self):
        # The float class has a is_integer method. Maybe Cents could extend float
        #if isinstance(self.value, int):
        if isinstance(self.value, int) or self.value.is_integer():
            return str(int(self.value)) + '¢'
        elif (self.value * 10).is_integer():
            #return ('%.1f¢') % self.real
            return str(self.real)
            # I think these are the same. Check Hz too
        else:
            return ('%.' + str(self.number_of_decimal_places) + 'f¢') % self.value

Intervals          = 'U22334T56677899AABBCDDEEF'
Interval_Qualities = 'PmMmMPTPmMmMPmMmMP#PmMmMP'

#class Interval_Quality:
#    def __init__(self, name)

class PitchBend:
    def __init__(self, cents):
        # round is strange, round(.5) returns 0 but round(.51) returns 1

        # Here -100 is 0 but 100 is 8192. You need to review where 0 pb is!

        #if cents < 100
        # You need to fix this too
        self.value = round(((8192 * cents.value/100) + 8192)/2)

    def lsb(self):
        return self.value & 127

    def msb(self):
        return self.value >> 7

class Interval:
    def __init__(self, a, b):
        # This is used for negating only
        self.a = a
        self.b = b

        self.midi_interval = abs(a.midi() - b.midi())
        # Call ^ n_semitones
        self.direction = int(self.midi_interval / (a.midi() - b.midi()))
        if self.midi_interval == 6:
            self.interval_base = 'T'
        else:
            self.interval_base = str(int(Intervals[self.midi_interval], 16))
        self.quality = Interval_Qualities[self.midi_interval]

    def __str__(self):
        return (self.direction*'Ascending ' + -self.direction*'Descending ') + self.quality + self.interval_base

    def __repr__(self):
        return -self.direction*'-' + self.quality + self.interval_base

    def __neg__(self):
        # Calculate Pitches and then give interval?
        return Interval(self.b, self.a)

def factor(n):
    if not isinstance(n, int) or n == 0 or n == 1:
        return []
    for i in range(2, int(sqrt(n))+1):
        if n % i == 0:
            return [i] + factor(int(n/i))
    return [n]

def is_prime(n):
    return len(factor(n)) == 1

primes = []

for i in range(2, 65536):
    if is_prime(i):
        primes += [i]

# Less resources is_prime
def is_prime(n):
    return n in primes


# float also has as_integer_ratio and is_integer
class Ratio:
    # Need a better name than numerator_or_str
    def __init__(self, numerator_or_str, denominator=1):
        if isinstance(numerator_or_str, Ratio):
            numerator, denominator = numerator_or_str.numerator, numerator_or_str.denominator
        if type(numerator_or_str) == float:
            # fractions
            import fractions
            f = fractions.Fraction(numerator_or_str).limit_denominator(1000)
            numerator, denominator = f.numerator, f.denominator
        elif type(numerator_or_str) == str:
            if numerator_or_str.count('/') == 1:
                numerator, denominator = [int(number_str) for number_str in numerator_or_str.split('/')]
            else:
                numerator = int(numerator_or_str)
        # This is for if the first or second arg is a Ratio. It's unfinished
        #if isinstance(numerator_or_str, self.__class__) or isinstance(numerator_or_str, self.__class__):
        #    if denominator == 1:
        #        denominator = Ratio(1)
        #    numerator = numerator_or_str.numerator * denom
        #if isinstance(denominator, self.__class__):
        elif isinstance(numerator_or_str, int):
            numerator = numerator_or_str
        if denominator == 0:
            raise ZeroDivisionError
        self.is_negative = ((numerator < 0) ^ (denominator < 0)) and (numerator != 0)
        self.numerator = abs(numerator)
        self.denominator = abs(denominator)

        # Reduce
        for prime in set(factor(self.numerator)) & set(factor(self.denominator)):
            while prime in set(factor(self.numerator)) and prime in set(factor(self.denominator)):
                self.numerator //= prime
                self.denominator //= prime

        self.numerator_factors = factor(self.numerator)
        self.denominator_factors = factor(self.denominator)

    def __eq__(self, other):
        return self.is_negative == other.is_negative and self.numerator == other.numerator and self.denominator == other.denominator

    def __gt__(self, other):
        return -(int(self.is_negative)*2-1) * self.numerator / self.denominator > \
            -(int(other.is_negative)*2-1) * other.numerator / other.denominator

    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)

    def __lt__(self, other):
        return -(int(self.is_negative)*2-1) * self.numerator / self.denominator < \
            -(int(other.is_negative)*2-1) * other.numerator / other.denominator

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __add__(self, other):
        if isinstance(other, int):
            return Ratio(-(int(self.is_negative)*2-1) * self.numerator + other * self.denominator, self.denominator)
        elif isinstance(other, Ratio):
            return Ratio(-(int(self.is_negative)*2-1) * self.numerator * other.denominator + -(int(other.is_negative)*2-1) * other.numerator * self.denominator, self.denominator * other.denominator)

    def __sub__(self, other):
        return self + (-other)

    def __neg__(self):
        if self.numerator != 0:
            self.is_negative = not self.is_negative
        return self

    def __mul__(self, other):
        if isinstance(other, int):
            return Ratio(self.numerator * other, self.denominator)
        elif isinstance(other, Ratio):
            return Ratio(self.numerator * other.numerator, self.denominator * other.denominator)

    def __truediv__(self, other):
        if isinstance(other, int):
            return Ratio(self.numerator, self.denominator * other)
        elif isinstance(other, Ratio):
            return Ratio(self.numerator * other.denominator, self.denominator * other.numerator)

    def __pow__(self, other):
        # Only ints for now. Ratios will come later
        if isinstance(other, int):
            r = Ratio(self.numerator, self.denominator)
            was_negative = r.is_negative
            if r.is_negative:
                r.is_negative = other % 2 == 1
            if other > 1:
                for i in range(other - 1):
                    r.numerator *= r.numerator
                    r.denominator *= r.denominator
            elif other == 1:
                pass
            elif other == 0:
                r.numerator = 1
                r.denominator = 1
            elif other < 0:
                old_numerator = r.numerator
                r.numerator = r.denominator
                r.denominator = old_numerator
                if other < -1:
                    for i in range(-other - 1):
                        r.numerator *= r.numerator
                        r.denominator *= r.denominator

            return r

    def __float__(self):
        return self.numerator / self.denominator

    def monzo(self):
        monzo_list = []

        if len(self.numerator_factors) != 0:
            numerator_monzo_len = primes.index(self.numerator_factors[-1])
        else:
            numerator_monzo_len = 0
        if len(self.denominator_factors) != 0:
            denominator_monzo_len = primes.index(self.denominator_factors[-1])
        else:
            denominator_monzo_len = 0

        if numerator_monzo_len > denominator_monzo_len:
            monzo_len = numerator_monzo_len
        else:
            monzo_len = denominator_monzo_len

        for i in range(monzo_len + 1):
            monzo_list.append(self.numerator_factors.count(primes[i]) - self.denominator_factors.count(primes[i]))
        return Monzo(monzo_list)

    def __repr__(self):
        return '{}{}/{}'.format('-'*self.is_negative, self.numerator, self.denominator)

    def factor_str(self):
        if len(self.numerator_factors) > 0:
            s = str(self.numerator_factors[0])
        else:
            s = '1'
        if len(self.numerator_factors) > 1:
            for n in self.numerator_factors[1:]:
                s += ' * ' + str(n)
        if len(self.denominator_factors) > 0:
            s += ' / ' + str(self.denominator_factors[0])
        if len(self.denominator_factors) > 1:
            for n in self.denominator_factors[1:]:
                s += ' / ' + str(n)
        return s

class Monzo:
    def __init__(self, monzo_list): # or_ratio?
        self.monzo_list = monzo_list

    def ratio(self):
        numerator = 1
        denominator = 1
        for i in range(len(self)):
            if self.monzo_list[i] > 0:
                numerator *= primes[i]**self.monzo_list[i]
            elif self.monzo_list[i] < 0:
                denominator *= primes[i]**-self.monzo_list[i]
        return Ratio(numerator, denominator)

    def to_list(self):
        return self.monzo_list

    def __getitem__(self, index):
        return self.monzo_list[index]

    def prime(self, prime_number):
        if prime_number in primes: # Or is_prime(pn). primes is made with the is_prime method now
            index = primes.index(prime_number)
            if index < len(self):
                return self.monzo_list[primes.index(prime_number)]
            else:
                return 0
        else:
            print('error')

    def __len__(self):
        return len(self.monzo_list)

    def __repr__(self):
        return '[{}⟩'.format(str(self.monzo_list)[1:-1])

def log2(x):
    return log(x) / log(2)

def test():
    print(Pitch(69))
    print(Pitch("A4"))
    print(Pitch("A4") == Pitch(69))
    #print(Pitch(69, -23) == Pitch(69) - Cents(23))
    #print(Pitch(69.1))
    A440 = Pitch(69)
    print(A440)
    C4 = Pitch(60)
    print(C4)
    print(A440 - C4)
    print(C4 - A440)
    interval = C4 - A440
    print(C4 + interval)
    print(C4 - interval)

    print(A440.midi(), A440.freq, A440.name())

    print('\n')

    print(C4 / 2)
    print(A440 * 5/4)
    print(A440 * 2)
    print(A440 * 3)
    print(Pitch(69, 2))

    print(Pitch(60-12*3))
    print(Pitch(60-12*3).get_freq())
    print(Pitch(60-12*3).midi_key)

    print(Pitch.from_name("A4").midi_key)
    print(Pitch(69) /16*27)

    r = Ratio(441,440)
    r.monzo()
    r.factor_str()
    Monzo([19,-12]).ratio()
    
    # These return False!
    Ratio(531441,524288).monzo() == Monzo([-19, 12])
    Ratio(2187, 2048) / Ratio(256, 243) == Ratio(531441, 524288)

    Pitch('A4') + Edostep(1, 12)
    Pitch('A4') + Edostep(1.1, 12)
    Pitch('A4') + Edostep(1, 41)
    Edostep(1, 12)
    Edostep(1.1, 12)

    Hz(440).to_midi_key()
    Pitch(Hz(440), 2)
    Pitch(Hz(440) + Cents(2))
    Pitch(Hz(441))
    Hz(440) + 100
    Hz(440) + Cents(100)

    # not working
    Pitch(69, 100)
    Pitch(69, -100)

if __name__ == '__main__':
    test()
